/* springmesh.c	-- Yasuhiro Yamazaki Tue Nov  6 2001
 */
#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "vector3d.h"
#include "pngtexture.h"

#define TRUE  (1)
#define FALSE (!TRUE)

/* ɥξ֤򼨤 */
#define WINDOW_STATE_SHOW 0
#define WINDOW_STATE_HIDE 1

/* ž ɥʬɥåȤɤž뤫򼨤 */
#define ROTATION_COEFFICIENT 180.0 /* degree */

/* åΥ */
#define MESH_WIDTH  15
#define MESH_HEIGHT 15

/*  */
#define GRAVITY -25.0f

/* Spring Mesh Structure */
typedef struct tagSpringLink
{
  /* ХͤĹȥХ */
  double length, power;
} SpringLink;

struct tagSpringPoint;

typedef struct tagSpringLinkList
{
  /* Ƹ³ꥹ */
  struct tagSpringPoint *point;
  SpringLink            *link;
  struct tagSpringLinkList *next;
} SpringLinkList;

typedef struct tagSpringPoint
{
  /*  */
  SpringLinkList *list;
  Vector3D position, velocity;
  Vector3D normal;
  double u, v;
  int fixed;
} SpringPoint;

/* Static ѿ */
static int windowState = WINDOW_STATE_SHOW;    /* ɥξ */
static double ratio_x = 0.0, ratio_y = 0.0;    /* ޥɸη׻ݻ */
static float rotate_x = 0.0f, rotate_y = 0.0f; /* ɽǥβž */
static int push_button;                        /* 줿ܥݻ */

static GLuint texture = 0; /* OpenGLѤƥID */
static SpringPoint *point[MESH_WIDTH][MESH_HEIGHT];  /* å */
static Vector3D gravity_src; /* ϥ٥ȥ */
static int toggle_normal = FALSE; /* ˡɽ̵ͭ */

/* Create Spring Structure */
static SpringLink *
createSpringLink( void )
{
  SpringLink *ret = (SpringLink *)malloc( sizeof( SpringLink ) );
  ret->length = 0.75;
  ret->power  = 800.0;
  return ret;
}
static SpringPoint *
createSpringPoint( void )
{
  SpringPoint *ret = (SpringPoint *)malloc( sizeof( SpringPoint ) );
  ret->list = NULL;
  ret->fixed = FALSE;
  return ret;
}

/* Spring link */
static void
addLink( SpringPoint *a, SpringLink *l, SpringPoint *b )
{
  SpringLinkList *add = (SpringLinkList *)malloc( sizeof( SpringLinkList ) );
  add->point = b;
  add->link  = l;
  add->next  = a->list;
  a->list    = add;
}
static void
linkBetween( SpringPoint *a, SpringLink *l, SpringPoint *b )
{
  addLink( a, l, b );
  addLink( b, l, a );
}

/*  */
static void
initializeAll( void )
{
  int x, y;
  SpringLink *link;
  
  /* ƥޤ */
  texture = createPngTexture( "sample.png" );

  /* åޤ */
  for( x=0; x<MESH_WIDTH; x++ ){
    for( y=0; y<MESH_HEIGHT; y++ ){
      point[x][y] = createSpringPoint();
    }
  }
  
  /* ³ꡢѥ᡼򥻥åȤޤ*/
  link = createSpringLink();
  for( x=0; x<MESH_WIDTH; x++ ){
    for( y=0; y<MESH_HEIGHT; y++ ){
      SpringPoint *a = point[x][y];
      if( x<MESH_WIDTH -1 ) linkBetween( a, link, point[x+1][y] );
      if( y<MESH_HEIGHT-1 ) linkBetween( a, link, point[x][y+1] );
      setVector3DCoordinatef( &a->position, 
			      (float)x-(float)(MESH_WIDTH-1)*0.5f,
			      (float)(MESH_HEIGHT-1-y)-(float)(MESH_HEIGHT-1)*0.5f,
			      0.0f );
      a->u = (double)x/(double)MESH_WIDTH;
      a->v = (double)y/(double)MESH_HEIGHT;
      setVector3DCoordinatef( &a->velocity,
			      0.0f, 0.0f, 0.0f );
      setVector3DCoordinatef( &a->normal,
			      1.0f, 0.0f, 0.0f );
    }
  }

  /* ꤹ롣 */
  point[               0][            0]->fixed = TRUE;
  point[(MESH_WIDTH-1)/2][MESH_HEIGHT-1]->fixed = TRUE;
  point[ MESH_WIDTH-1   ][            0]->fixed = TRUE;
}

/*  */
static void
freeAll( void )
{
  /* ƥ˴ޤ */
  destroyPngTexture( texture );
}

/* ϥ٥ȥη׻ */
static void
calculateGravity( void )
{
  float x, y, z;
  GLfloat m[16];

  x = 0.0f;
  y = GRAVITY;
  z = 0.0f;
  glGetFloatv( GL_MODELVIEW_MATRIX, m );
  gravity_src.x = m[ 0]*x+m[ 4]*y+m[ 8]*z;
  gravity_src.y = m[ 1]*x+m[ 5]*y+m[ 9]*z;
  gravity_src.z = m[ 2]*x+m[ 6]*y+m[10]*z;
}

/* ץ󥰤ν */
static void
updateSpringMesh( const double sec )
{
  int x, y, i;
  SpringPoint *p;
  SpringLinkList *here, *next;
  Vector3D strength, gravity, n, tmp, tmp2;
  float normal, length;
  
  copyVector3D( &gravity, &gravity_src );
  multiplyVector3D( &gravity, sec );

  /* ϴطη׻ */
  for( x=0; x<MESH_WIDTH; x++ ){
    for( y=0; y<MESH_HEIGHT; y++ ){
      p = point[x][y]; /* åθ */
      setVector3DCoordinatef( &strength, 0.0f, 0.0f, 0.0f );
      setVector3DCoordinatef( &n,        0.0f, 0.0f, 0.0f );
      here = p->list;
      i = 0;
      while( here!=NULL ){ /* Ϣ뤹Ʒ׻ */
	next = here->next;
	if( next==NULL ) next = p->list;

	/* ˡη׻ */
	subVector3D( &tmp, &here->point->position, &p->position );
	subVector3D( &tmp2, &next->point->position, &p->position );
	crossProductVector3D( &tmp2, &tmp, &tmp2 );
	normalizeVector3D( &tmp2 );
	addVector3D( &n, &n, &tmp2 );

	/* ĥ */
	subVector3D( &tmp, &here->point->position, &p->position );
	normal = getNormal( &tmp );
	length = normal - here->link->length; /*  */
	divideVector3D( &tmp, normal ); /* ñ̥٥ȥ벽 */

	multiplyVector3D( &tmp, length * here->link->power ); /* ĥ */
	addVector3D( &strength, &strength, &tmp ); /*  */
	
	here = here->next;
	i++;
      }
      divideVector3D( &n, (float)i );
      copyVector3D( &p->normal, &n );

      /* ®٤ꤹ */
      multiplyVector3D( &strength, sec );
      addVector3D( &p->velocity, &p->velocity, &strength ); /* Ϥû */
      addVector3D( &p->velocity, &p->velocity, &gravity  ); /* Ϥû */
    }
  }

  /* Ϥη׻̤ƤθŬ롣 */
  for( x=0; x<MESH_WIDTH; x++ ){
    for( y=0; y<MESH_HEIGHT; y++ ){
      p = point[x][y];
      if( !p->fixed ){
	copyVector3D( &tmp, &p->velocity );
	multiplyVector3D( &tmp, sec );
	addVector3D( &p->position, &p->position, &tmp );
	multiplyVector3D( &p->velocity, 0.95f );
      }
    }
  }
}

/* ɽ */
static void 
putPlane( void )
{
  GLfloat light_position[]    = { 0.0f, 0.0f, -1.0f, 0.0f };
  GLfloat light_color[]       = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat light_ambient[]     = { 0.0f, 0.0f, 0.0f, 1.0f };
  GLfloat material_diffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat material_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat material_ambient[]  = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat material_emission[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  int           x,   y;
  SpringPoint *p0, *p1;
  Vector3D    *v0, *v1;
  Vector3D    *n0, *n1;
  
  glPushMatrix();
    glRotatef( -rotate_x, 1.0f, 0.0f, 0.0f);
    glRotatef( -rotate_y, 0.0f, 1.0f, 0.0f);
    calculateGravity();
  glPopMatrix();

  glPushMatrix();
    /* 饤Ȥ */
    glLightfv( GL_LIGHT0, GL_POSITION, light_position );
    glLightfv( GL_LIGHT0, GL_DIFFUSE,     light_color );
    glLightfv( GL_LIGHT0, GL_SPECULAR,    light_color );
    glLightfv( GL_LIGHT0, GL_AMBIENT,   light_ambient );
    
    /* 饤Ȥ򥪥 */
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    
    /* ž */
    glRotatef( rotate_y, 0.0f, 1.0f, 0.0f);
    glRotatef( rotate_x, 1.0f, 0.0f, 0.0f);
    glColor3f( 1.0f, 1.0f, 1.0f );

    /* ޥƥꥢ */
    glMaterialfv( GL_FRONT, GL_DIFFUSE,    material_diffuse );
    glMaterialfv( GL_FRONT, GL_SPECULAR,  material_specular );
    glMaterialfv( GL_FRONT, GL_EMISSION,  material_emission );
    glMaterialfv( GL_FRONT, GL_AMBIENT,    material_ambient );
    glMaterialf ( GL_FRONT, GL_SHININESS,            120.0f );

    glMaterialfv( GL_BACK, GL_DIFFUSE,    material_diffuse );
    glMaterialfv( GL_BACK, GL_SPECULAR,  material_specular );
    glMaterialfv( GL_BACK, GL_EMISSION,  material_emission );
    glMaterialfv( GL_BACK, GL_AMBIENT,    material_ambient );
    glMaterialf ( GL_BACK, GL_SHININESS,            120.0f );

    /* ƥͭˤޤ */
    glEnable( GL_TEXTURE_2D );
 
    /* Ѥƥꤷޤ */
    glBindTexture( GL_TEXTURE_2D, texture );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

    /* Ķޥåԥ */
    glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
    glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    glEnable( GL_TEXTURE_GEN_S );
    glEnable( GL_TEXTURE_GEN_T );

    /* å */
    for( x=0; x<MESH_WIDTH-1; x++ ){
      /* Ӿ */
      glBegin( GL_TRIANGLE_STRIP );
      for( y=0; y<MESH_HEIGHT; y++ ){
	p0 = point[x  ][y];
	p1 = point[x+1][y];
	v0 = &p0->position;
	v1 = &p1->position;
	n0 = &p0->normal;
	n1 = &p1->normal;
	glTexCoord2f( p0->u, p0->v ); glNormal3f( n0->x, n0->y, n0->z );
	glVertex3f( v0->x, v0->y, v0->z );
	glTexCoord2f( p1->u, p1->v ); glNormal3f( n1->x, n1->y, n1->z );
	glVertex3f( v1->x, v1->y, v1->z );
      }
      glEnd();
    }

    /* Ķޥåԥ󥰤̵ˤޤ */
    glDisable( GL_TEXTURE_GEN_T );
    glDisable( GL_TEXTURE_GEN_S );

    /* ƥ̵ˤޤ */
    glDisable( GL_TEXTURE_2D );

    /* 饤Ȥ򥪥 */
    glDisable( GL_LIGHT0 );
    glDisable( GL_LIGHTING );

    /* ˡ */
    if( toggle_normal ){
      glColor3f( 1.0f, 1.0f, 0.0f );
      glBegin( GL_LINES );
      for( x=0; x<MESH_WIDTH; x++ ){
	for( y=0; y<MESH_HEIGHT; y++ ){
	  p0 = point[x][y];
	  v0 = &p0->position;
	  v1 = &p0->normal;
	  glVertex3f( v0->x, v0->y, v0->z );
	  glVertex3f( v0->x+v1->x, v0->y+v1->y, v0->z+v1->z );
	}
      }
      glEnd();
    }

    /* Ȥ */
    glColor3f( 0.0f, 1.0f, 0.0f );
    glBegin( GL_LINE_LOOP );
    glVertex3f( -(float)(MESH_WIDTH-1)*0.5f, -(float)(MESH_HEIGHT-1)*0.5f, 0.0f );
    glVertex3f(  (float)(MESH_WIDTH-1)*0.5f, -(float)(MESH_HEIGHT-1)*0.5f, 0.0f );
    glVertex3f(  (float)(MESH_WIDTH-1)*0.5f,  (float)(MESH_HEIGHT-1)*0.5f, 0.0f );
    glVertex3f( -(float)(MESH_WIDTH-1)*0.5f,  (float)(MESH_HEIGHT-1)*0.5f, 0.0f );
    glEnd();
    
  glPopMatrix();
}

/* 褬ɬפȤʤä˸ƤФ */
static void
displayCallback( void )
{
  glLoadIdentity();
  glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT );

  gluLookAt( 0.0, -5.0, 12.5, 0.0, -5.0, 0.0, 0.0, 1.0, 0.0 );

  /* Put something */
  putPlane();
  
  glutSwapBuffers();
}

/* ɥꥵ줿˸ƤФ */
static void
reshapeCallback( int width, int height )
{
  glViewport( 0, 0, width, height );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  gluPerspective( 60.0, width/(GLdouble)height, 0.01, 100 );

  glMatrixMode( GL_MODELVIEW );
}

/* ɥϾƤФ */
static void
idleCallback( void )
{
  if( windowState==WINDOW_STATE_SHOW ){
    /* ɥ˹Ԥ򤳤˵ */
    updateSpringMesh( 0.025 );
    glutPostRedisplay();
  }
}

/* ɥ{ɽ/줿}˸ƤФ */
static void
visibilityCallback( int state )
{
  switch( state ){
  case GLUT_NOT_VISIBLE:
    fprintf( stderr, "THE WINDOW WAS HIDDEN.\n" );
    windowState = WINDOW_STATE_HIDE;
    break;
  case GLUT_VISIBLE:
    fprintf( stderr, "THE WINDOW WAS SHOWN.\n" );
    windowState = WINDOW_STATE_SHOW;
    break;
  }
}

/* ޥɸΥɥФ׻ */
static void
calculationRatioOfMouseCoordinate( const int x, const int y )
{
  int width, height;
  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );
  ratio_x = (double)y/(double)height;
  ratio_y = (double)x/(double)width;
}

/* ޥΥܥѲä˸ƤФ */
static void
mouseCallback( int button, int state, int x, int y )
{
  switch( button ){
  case GLUT_LEFT_BUTTON:
    push_button = GLUT_LEFT_BUTTON;
    fprintf( stderr, "LEFT BUTTON: " );
    break;
  case GLUT_MIDDLE_BUTTON:
    push_button = GLUT_MIDDLE_BUTTON;
    fprintf( stderr, "MIDDLE BUTTON: " );
    break;
  case GLUT_RIGHT_BUTTON:
    push_button = GLUT_RIGHT_BUTTON;
    fprintf( stderr, "RIGHT BUTTON: " );
    break;
  }
  
  switch( state ){
  case GLUT_DOWN:
    fprintf( stderr, "DOWN\n" );
    break;
  case GLUT_UP:
    fprintf( stderr, "UP\n" );
    break;
  }

  calculationRatioOfMouseCoordinate( x, y );
}

/* ޥΥݥ󥿤ưä˸ƤФ */
static void
passiveMotionCallback( int x, int y )
{
  fprintf( stderr, "MOUSE (%d, %d)\n", x, y );
}

/* ɥå桢ޥΥݥ󥿤ưä˸ƤФ */
static void
motionCallback( int x, int y )
{
  double rx, ry;

  fprintf( stderr, "DRAG (%d, %d) ", x, y );

  switch( push_button ){ /* Ƥܥ˱ȿ򤵤 */
  case GLUT_LEFT_BUTTON:
    fprintf( stderr, "LEFT BUTTON\n" );
    /* ʪΤβž׻ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    rotate_x += (float)((ratio_x-rx)*ROTATION_COEFFICIENT);
    rotate_y += (float)((ratio_y-ry)*ROTATION_COEFFICIENT);
    glutPostRedisplay();
    break;
  case GLUT_MIDDLE_BUTTON:
    fprintf( stderr, "MIDDLE BUTTON\n" );
    break;
  case GLUT_RIGHT_BUTTON:
    fprintf( stderr, "RIGHT BUTTON\n" );
    break;
  }
}

/* Ϥä˸ƤФ */
static void
keyboardCallback( unsigned char key, int x, int y )
{
  fprintf( stderr, "HIT \'%c\' KEY (CODE: %d)\n", (key<32)?'?':key, (int)key );

  switch( key ){
  /*
  ** Esc줿ץλ
  */
  case 27:
    exit( EXIT_SUCCESS );
    break;

  case 'n':
    if( toggle_normal ) toggle_normal = FALSE;
    else toggle_normal = TRUE;
    glutPostRedisplay();
    break;
    
  default:
    break;
  }
}

/*  */
static void
initGL( void )
{
  int width, height;

  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );

  reshapeCallback( width, height );

  glClearColor( 0.25, 0.25, 0.25, 1.0 );
  glEnable( GL_DEPTH_TEST );
  glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
}

/* MAIN */
int
main( int argc, char *argv[] )
{
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH );

  glutCreateWindow( "Demo: TEXTURE ANIMATION" );

  glutDisplayFunc      ( displayCallback       );
  glutKeyboardFunc     ( keyboardCallback      );
  glutReshapeFunc      ( reshapeCallback       );
  glutIdleFunc         ( idleCallback          );
  glutVisibilityFunc   ( visibilityCallback    );
  glutMouseFunc        ( mouseCallback         );
  glutPassiveMotionFunc( passiveMotionCallback );
  glutMotionFunc       ( motionCallback        );
  
  initGL();
  initializeAll();
  
  glutMainLoop();

  freeAll();
  
  return EXIT_SUCCESS;
}
